@@ -36,6 +36,7 @@
 #include "muse_flux.h"
 #include "muse_imagelist.h"
 #include "muse_pfits.h"
+#include "muse_phys.h"
 #include "muse_rvcorrect.h"
 #include "muse_utils.h"
 #include "muse_wcs.h"
@@ -45,6 +46,11 @@
 #include "muse_sky_old.h"
 #endif
 
+#undef CREATE_IDP
+#ifdef CREATE_IDP
+#include "muse_idp.h"
+#endif
+
 /*----------------------------------------------------------------------------*/
 /**
  * @defgroup muse_postproc       Post-processing functions
@@ -266,6 +272,9 @@ muse_postproc_get_cube_format(const char *aFormatString)
   if (!strncmp(aFormatString, "xEuro3D", 8)) {
     return MUSE_CUBE_TYPE_EURO3D_X;
   }
+  if (!strncmp(aFormatString, "sdpCube", 8)) {
+    return MUSE_CUBE_TYPE_SDP;
+  }
   return MUSE_CUBE_TYPE_FITS; /* catch all fallback */
 } /* muse_postproc_get_cube_format() */
 
@@ -536,19 +545,12 @@ muse_postproc_process_exposure(muse_postproc_properties *aProp,
                                                   aProp->skymodel_params.fraction,
             /* XXX hardcode prefix for now... */  "ESO QC SCIPOST");
 
-    /* Apply mask and create spectrum */
+    /* Apply mask and create spectrum, possibly with one CR-reject iteration */
     cpl_table_select_all(sky_pt->table);
     muse_pixtable_and_selected_mask(sky_pt, sky_mask);
-    cpl_table *spectrum = muse_resampling_spectrum(sky_pt,
-                                                   aProp->skymodel_params.sampling);
-    if (aProp->skymodel_params.crsigmas > 0.) {
-      cpl_msg_info(__func__, "Recreating sky spectrum with %.3f sigma level CR "
-                   "rejection", aProp->skymodel_params.crsigmas);
-      muse_sky_mark_cosmic(spectrum, sky_pt, aProp->skymodel_params.crsigmas);
-      cpl_table_delete(spectrum);
-      spectrum = muse_resampling_spectrum(sky_pt, aProp->skymodel_params.sampling);
-    } /* if aProp->skymodel_params.crsigmas positive */
-
+    cpl_table *spectrum =
+      muse_resampling_spectrum_iterate(sky_pt, aProp->skymodel_params.sampling,
+                                       0., aProp->skymodel_params.crsigmas, 1);
     if (aProp->skymethod == MUSE_POSTPROC_SKYMETHOD_MODEL) {
       /* Fit sky lines spectrum */
       cpl_array *lambda = muse_cpltable_extract_column(spectrum, "lambda");
@@ -559,7 +561,7 @@ muse_postproc_process_exposure(muse_postproc_properties *aProp,
 
       // do the fit, ignoring possible errors
       prestate = cpl_errorstate_get();
-      if (aProp->lsf_cube != NULL) {
+      if (aProp->lsf_cube) {
         cpl_image *lsfImage = muse_lsf_average_cube_all(aProp->lsf_cube, sky_pt);
         muse_wcs *lsfWCS = muse_lsf_cube_get_wcs_all(aProp->lsf_cube);
         /* Filter the image with a rectangle to model the spectrum binning */
@@ -572,7 +574,7 @@ muse_postproc_process_exposure(muse_postproc_properties *aProp,
         }
 
         /* Create continuum */
-        if (aProp->sky_continuum == NULL) {
+        if (!aProp->sky_continuum) {
           cpl_msg_info(__func__, "No sky continuum given, create it from the data");
           sky_cont = muse_sky_continuum_create(spectrum, sky_lines, lsfImage, lsfWCS,
                                                aProp->skymodel_params.csampling);
@@ -583,7 +585,7 @@ muse_postproc_process_exposure(muse_postproc_properties *aProp,
         /* clean up */
         cpl_image_delete(lsfImage);
 #ifdef USE_LSF_PARAMS
-      } else if (aProp->lsf_params != NULL) {// Old LSF params code
+      } else if (aProp->lsf_params) { // Old LSF params code
         // do the fit, ignoring possible errors
         muse_sky_lines_fit_old(spectrum, sky_lines);
         if (!cpl_errorstate_is_equal(prestate)) {
@@ -591,14 +593,14 @@ muse_postproc_process_exposure(muse_postproc_properties *aProp,
           cpl_errorstate_set(prestate);
         }
 
-        if (aProp->sky_continuum == NULL) {
+        if (!aProp->sky_continuum) {
           cpl_msg_info(__func__, "No sky continuum given, create it from the data");
           muse_sky_subtract_lines_old(sky_pt, sky_lines, aProp->lsf_params);
-          aProp->sky_continuum
-            = muse_resampling_spectrum(sky_pt, aProp->skymodel_params.csampling);
-          cpl_table_erase_column(aProp->sky_continuum, "stat");
-          cpl_table_erase_column(aProp->sky_continuum, "dq");
-          cpl_table_name_column(aProp->sky_continuum, "data", "flux");
+          sky_cont = muse_resampling_spectrum(sky_pt,
+                                              aProp->skymodel_params.csampling);
+          cpl_table_erase_column(sky_cont, "stat");
+          cpl_table_erase_column(sky_cont, "dq");
+          cpl_table_name_column(sky_cont, "data", "flux");
         }
 #endif
       }
@@ -667,22 +669,6 @@ muse_postproc_process_exposure(muse_postproc_properties *aProp,
     cpl_table_delete(sky_lines);
     cpl_table_delete(sky_cont);
   }
-  if (aProp->skymethod == MUSE_POSTPROC_SKYMETHOD_ROWBYROW) {
-    /* apply the row-by-row sky subtraction; needed for row-by-row */
-    cpl_msg_debug(__func__, "doing sky subtraction using row-by-row fitting");
-    /* convert pixel table into images, one per IFU */
-    muse_imagelist *list = muse_pixtable_to_imagelist(pt);
-    unsigned int k, n = muse_imagelist_get_size(list);
-    #pragma omp parallel for default(none)                                     \
-            shared(aProp, list, n)
-    for (k = 0; k < n; k++) {
-      muse_image *image = muse_imagelist_get(list, k);
-      muse_sky_subtract_rowbyrow_mask(image, NULL);
-      muse_sky_subtract_rowbyrow(image, NULL, 3., 4);
-    } /* for all IFUs / images */
-    muse_pixtable_from_imagelist(pt, list);
-    muse_imagelist_delete(list);
-  } /* if row-by-row sky subtraction */
 
   /* carry out radial velocity correction, if specified */
   muse_rvcorrect(pt, aProp->rvtype);
@@ -753,9 +739,24 @@ muse_postproc_process_exposure(muse_postproc_properties *aProp,
     muse_wcs_object *wcs = muse_wcs_object_new();
     wcs->crpix1 = aProp->crpix1; /* set center of rotation */
     wcs->crpix2 = aProp->crpix2;
-    rc = muse_wcs_locate_sources(pt, aProp->detsigma, aProp->centroid, wcs);
-    rc = muse_wcs_solve(wcs, reftable, aProp->radius, aProp->faccuracy,
-                        aProp->niter, aProp->rejsigma);
+    rc = muse_wcs_locate_sources(pt, fabsf(aProp->detsigma), aProp->centroid,
+                                 wcs);
+    if (aProp->detsigma < 0.) {
+      /* loop through possible detection sigmas to find the optimal solution */
+      rc = muse_wcs_optimize_solution(wcs, aProp->detsigma, aProp->centroid,
+                                      reftable, aProp->radius, aProp->faccuracy,
+                                      aProp->niter, aProp->rejsigma);
+    } else {
+      /* use just the given detection sigma for the solution */
+      rc = muse_wcs_solve(wcs, reftable, aProp->radius, aProp->faccuracy,
+                          aProp->niter, aProp->rejsigma);
+      if (wcs->wcs) { /* set the keyword that muse_wcs_solve() does not save */
+        cpl_propertylist_update_float(wcs->wcs, MUSE_HDR_WCS_DETSIGMA,
+                                      aProp->detsigma);
+        cpl_propertylist_set_comment(wcs->wcs, MUSE_HDR_WCS_DETSIGMA,
+                                     MUSE_HDR_WCS_DETSIGMA_C_ONE);
+      }
+    }
     cpl_table_delete(reftable);
     cpl_msg_debug(__func__, "Solving astrometry (initial radius %.1f pix, "
                   "initial relative accuracy %.1f, %d iterations with rejection"
@@ -842,7 +843,8 @@ muse_postproc_cube_load_output_wcs(muse_processing *aProcessing)
     if (valid && cpl_array_get_string(ctypes, 2)) {
       const char *ctype3 = cpl_array_get_string(ctypes, 2);
       valid = valid
-            && (!strncmp(ctype3, "AWAV", 5) || !strncmp(ctype3, "AWAV-LOG", 9));
+            && (!strncmp(ctype3, "AWAV", 5) || !strncmp(ctype3, "AWAV-LOG", 9) ||
+                !strncmp(ctype3, "WAVE", 5) || !strncmp(ctype3, "WAVE-LOG", 9));
     }
     if (valid) { /* check for unsupported PCi_j keywords */
       /* still needs to be done on the input header not on the derives WCS! */
@@ -899,7 +901,15 @@ muse_postproc_cube_resample_and_collapse(muse_processing *aProcessing,
   cpl_ensure_code(aFormat == MUSE_CUBE_TYPE_EURO3D ||
                   aFormat == MUSE_CUBE_TYPE_FITS ||
                   aFormat == MUSE_CUBE_TYPE_EURO3D_X ||
-                  aFormat == MUSE_CUBE_TYPE_FITS_X, CPL_ERROR_ILLEGAL_INPUT);
+                  aFormat == MUSE_CUBE_TYPE_FITS_X ||
+                  aFormat == MUSE_CUBE_TYPE_SDP, CPL_ERROR_ILLEGAL_INPUT);
+
+  /* convert cube from air to vacuum wavelengths, if needed */
+  if (aParams->tlambda == MUSE_RESAMPLING_DISP_WAVE ||
+      aParams->tlambda == MUSE_RESAMPLING_DISP_WAVE_LOG) {
+    /* convert using standard air (and Ciddor method) */
+    muse_phys_air_to_vacuum(aPixtable, MUSE_PHYS_AIR_STANDARD);
+  }
 
   /* create cube and cast to generic pointer to save code duplication */
   void *cube = NULL;
@@ -912,7 +922,8 @@ muse_postproc_cube_resample_and_collapse(muse_processing *aProcessing,
     cpl_msg_indent_less();
     cpl_ensure_code(e3d, cpl_error_get_code());
     cube = e3d;
-  } else if (aFormat == MUSE_CUBE_TYPE_FITS || aFormat == MUSE_CUBE_TYPE_FITS_X) {
+  } else if (aFormat == MUSE_CUBE_TYPE_FITS ||
+      aFormat == MUSE_CUBE_TYPE_FITS_X || aFormat == MUSE_CUBE_TYPE_SDP) {
     cpl_msg_info(__func__, "Resampling to final cube follows, output format "
                  "\"FITS cube\"");
     cpl_msg_indent_more();
@@ -937,7 +948,8 @@ muse_postproc_cube_resample_and_collapse(muse_processing *aProcessing,
     muse_image *fov = NULL;
     if (aFormat == MUSE_CUBE_TYPE_EURO3D || aFormat == MUSE_CUBE_TYPE_EURO3D_X) {
       fov = muse_euro3dcube_collapse(cube, filter);
-    } else if (aFormat == MUSE_CUBE_TYPE_FITS || aFormat == MUSE_CUBE_TYPE_FITS_X) {
+    } else if (aFormat == MUSE_CUBE_TYPE_FITS ||
+        aFormat == MUSE_CUBE_TYPE_FITS_X || aFormat == MUSE_CUBE_TYPE_SDP) {
       if (getenv("MUSE_COLLAPSE_PIXTABLE") &&
           atoi(getenv("MUSE_COLLAPSE_PIXTABLE")) > 0) {
         /* create collapsed image directly from pixel table/grid */
@@ -978,20 +990,41 @@ muse_postproc_cube_resample_and_collapse(muse_processing *aProcessing,
   cpl_array_delete(filters);
   muse_pixgrid_delete(grid);
 
+#ifdef CREATE_IDP
+  /* Collect IDP properties from the raw data frames and the data cube *
+   * header                                                            */
+  muse_idp_properties *properties =
+      muse_idp_properties_collect(aProcessing, ((muse_datacube *)cube)->header,
+                                  MUSE_TAG_CUBE_FINAL);
+
+  /* Write formatted IDP properties to the data cube product header so *
+   * that they are saved to the product file in the next step          */
+  muse_idp_properties_update((muse_datacube *)cube)->header, properties);
+#endif
+
   cpl_error_code rc = CPL_ERROR_NONE;
   if (aFormat == MUSE_CUBE_TYPE_EURO3D || aFormat == MUSE_CUBE_TYPE_EURO3D_X) {
     /* save the data and clean up */
     rc = muse_processing_save_cube(aProcessing, -1, cube, MUSE_TAG_CUBE_FINAL,
                                    MUSE_CUBE_TYPE_EURO3D);
     muse_euro3dcube_delete(cube);
-  } else if (aFormat == MUSE_CUBE_TYPE_FITS || aFormat == MUSE_CUBE_TYPE_FITS_X) {
+  } else if (aFormat == MUSE_CUBE_TYPE_FITS || aFormat == MUSE_CUBE_TYPE_FITS_X ||
+      aFormat == MUSE_CUBE_TYPE_SDP) {
     /* convert/remove the DQ cube */
     muse_datacube_convert_dq(cube);
     /* save the cube and clean up */
-    rc = muse_processing_save_cube(aProcessing, -1, cube, MUSE_TAG_CUBE_FINAL,
+    if (aFormat == MUSE_CUBE_TYPE_SDP) {
+      rc = muse_processing_save_cube(aProcessing, -1, cube, MUSE_TAG_CUBE_FINAL,
+                                   MUSE_CUBE_TYPE_SDP);
+    } else {
+      rc = muse_processing_save_cube(aProcessing, -1, cube, MUSE_TAG_CUBE_FINAL,
                                    MUSE_CUBE_TYPE_FITS);
+    }
     muse_datacube_delete(cube);
   }
+#ifdef CREATE_IDP
+  muse_idp_properties_delete(properties);
+#endif
   return rc;
 } /* muse_postproc_cube_resample_and_collapse() */
 
@@ -1125,9 +1158,11 @@ muse_postproc_qc_fwhm(muse_processing *aProcessing, muse_datacube *aCube)
     cpl_size xcen = lround(cpl_apertures_get_centroid_x(apertures, n)),
              ycen = lround(cpl_apertures_get_centroid_y(apertures, n));
     double x, y;
+    cpl_errorstate state = cpl_errorstate_get();
     cpl_image_get_fwhm(cim, xcen, ycen, &x, &y);
     const char *fwhmcomment = NULL;
-    if (x < 0 || y < 0) { /* something's weird with this source */
+    if (x < 0 || y < 0 || !cpl_errorstate_is_equal(state)) {
+      /* something's weird with this source */
       x = -1.;
       y = -1.;
       /* do not count these in the statistics */
@@ -1135,6 +1170,8 @@ muse_postproc_qc_fwhm(muse_processing *aProcessing, muse_datacube *aCube)
       cpl_image_reject(statimage, n, 2);
       fwhmcomment = "[arcsec] failure determining FWHM";
       nbad++;
+      /* suppress the error state */
+      cpl_errorstate_set(state);
     } else {
       /* linear WCS scaling */
       x = cd11 * x + cd12 * y;
@@ -1170,16 +1207,19 @@ muse_postproc_qc_fwhm(muse_processing *aProcessing, muse_datacube *aCube)
   /* add simple FWHM statistics as well */
   snprintf(keyword, KEYWORD_LENGTH, QC_POST_FWHM_NVALID, prefix);
   cpl_propertylist_update_int(aCube->header, keyword, ndet - nbad);
+  cpl_errorstate state = cpl_errorstate_get();
   cpl_stats *s = cpl_stats_new_from_image(statimage,
                                           CPL_STATS_MEDIAN | CPL_STATS_MAD);
+  cpl_boolean error = cpl_errorstate_is_equal(state);
+  cpl_errorstate_set(state); /* swallow the error again */
   snprintf(keyword, KEYWORD_LENGTH, QC_POST_FWHM_MEDIAN, prefix);
-  if (ndet < 3) {
+  if (error || ndet < 3) {
     cpl_propertylist_update_float(aCube->header, keyword, 0.);
   } else {
     cpl_propertylist_update_float(aCube->header, keyword, cpl_stats_get_median(s));
   }
   snprintf(keyword, KEYWORD_LENGTH, QC_POST_FWHM_MAD, prefix);
-  if (ndet < 3) {
+  if (error || ndet < 3) {
     cpl_propertylist_update_float(aCube->header, keyword, 0.);
   } else {
     cpl_propertylist_update_float(aCube->header, keyword, cpl_stats_get_mad(s));
