@@ -40,6 +40,7 @@
 #include "muse_lsf.h"
 #include "muse_datacube.h"
 #include "muse_utils.h"
+#include "muse_idp.h"
 
 /*----------------------------------------------------------------------------*
  *                             Debugging Macros                               *
@@ -314,7 +315,8 @@ muse_processing_setup_header(muse_processing *aProcessing,
   cpl_frameset *fset = muse_frameset_sort_raw_other(aProcessing->usedframes,
                                                     aIndex, aDateObs, aSequence);
 
-  /* XXX workaround for the EQUINOX warning, replace with double-type keyword */
+#if CPL_VERSION_CODE < CPL_VERSION(7, 0, 0)
+  /* workaround for the EQUINOX warning, replace with double-type keyword */
   if (cpl_propertylist_has(aHeader, "EQUINOX") &&
       cpl_propertylist_get_type(aHeader, "EQUINOX") < CPL_TYPE_FLOAT) {
     cpl_property *equinox =  cpl_propertylist_get_property(aHeader, "EQUINOX");
@@ -326,6 +328,7 @@ muse_processing_setup_header(muse_processing *aProcessing,
     cpl_propertylist_set_comment(aHeader, "EQUINOX", equcomment);
     cpl_propertylist_erase(aHeader, "EQUIBRK");
   } /* if non-float EQUINOX */
+#endif
 
   char *pkgstring = cpl_sprintf("%s/%s", PACKAGE, PACKAGE_VERSION);
   if (cpl_dfs_setup_product_header(aHeader, aFrame, fset,
@@ -490,7 +493,8 @@ muse_processing_save_image(muse_processing *aProcessing, int aIFU,
   @param  aCube         the cube to be saved
   @param  aTag          tag to apply for output frames, used as default prefix
   @param  aType         type of cube to save (MUSE_CUBE_TYPE_FITS,
-                        MUSE_CUBE_TYPE_EURO3D, or MUSE_CUBE_TYPE_LSF)
+                        MUSE_CUBE_TYPE_EURO3D, MUSE_CUBE_TYPE_SDP, or
+                        MUSE_CUBE_TYPE_LSF)
   @return CPL_ERROR_NONE or the relevant cpl_error_code on error
 
   Save a datacube to disk. The filename consists on the prefix, a "-"
@@ -507,14 +511,34 @@ muse_processing_save_cube(muse_processing *aProcessing, int aIFU,
                           void *aCube, const char *aTag, muse_cube_type aType)
 {
   cpl_ensure_code(aProcessing && aCube && aTag, CPL_ERROR_NULL_INPUT);
-  cpl_ensure_code(aType == MUSE_CUBE_TYPE_EURO3D || aType == MUSE_CUBE_TYPE_FITS ||
-                  aType == MUSE_CUBE_TYPE_LSF, CPL_ERROR_ILLEGAL_INPUT);
+  cpl_ensure_code((aType == MUSE_CUBE_TYPE_EURO3D) ||
+                  (aType == MUSE_CUBE_TYPE_FITS) ||
+                  (aType == MUSE_CUBE_TYPE_SDP) ||
+                  (aType == MUSE_CUBE_TYPE_LSF), CPL_ERROR_ILLEGAL_INPUT);
 
   /* both cube structures contain the header component as  *
    * first element, so it's safe to cast it to one of them */
   cpl_frame *frame = muse_processing_new_frame(aProcessing, aIFU,
                                                ((muse_datacube *)aCube)->header,
                                                aTag, CPL_FRAME_TYPE_IMAGE);
+
+  /* For science data cubes collect IDP properties from the raw data frames *
+   * and the data cube header, and update the data cube product header so   *
+   * that they are saved in the next step.                                  */
+  if (aType == MUSE_CUBE_TYPE_SDP) {
+    cpl_errorstate status = cpl_errorstate_get();
+    muse_idp_properties *properties =
+        muse_idp_properties_collect(aProcessing, (muse_datacube *)aCube, aTag);
+    if (properties) {
+      muse_idp_properties_update(((muse_datacube *)aCube)->header, properties);
+    }
+    muse_idp_properties_delete(properties);
+    if (cpl_errorstate_get() != status) {
+      cpl_frame_delete(frame);
+      return cpl_error_get_code();
+    }
+  }
+
   cpl_msg_info(__func__, "Saving %s cube as \"%s\"",
                aType == MUSE_CUBE_TYPE_EURO3D ? "Euro3D" : "FITS",
                cpl_frame_get_filename(frame));
@@ -522,7 +546,7 @@ muse_processing_save_cube(muse_processing *aProcessing, int aIFU,
   if (aType == MUSE_CUBE_TYPE_EURO3D) {
     rc = muse_euro3dcube_save((muse_euro3dcube *)aCube,
                               cpl_frame_get_filename(frame));
-  } else if (aType == MUSE_CUBE_TYPE_FITS) {
+  } else if ((aType == MUSE_CUBE_TYPE_FITS) || (aType == MUSE_CUBE_TYPE_SDP)) {
     rc = muse_datacube_save((muse_datacube *)aCube,
                             cpl_frame_get_filename(frame));
   } else {
@@ -583,8 +607,12 @@ muse_processing_save_table(muse_processing *aProcessing, int aIFU,
                cpl_frame_get_filename(frame));
   cpl_error_code rc = CPL_ERROR_NONE;
   if (aType == MUSE_TABLE_TYPE_CPL) {
+    cpl_size nrow = cpl_table_get_nrow((cpl_table *)aTable);
     rc = cpl_table_save((cpl_table *)aTable, aHeader, NULL,
                         cpl_frame_get_filename(frame), CPL_IO_CREATE);
+    if (nrow < 1) {
+      cpl_msg_warning(__func__, "Table saved as \"%s\" has no rows!", aTag);
+    }
   } else {
     rc = muse_pixtable_save((muse_pixtable *)aTable,
                             cpl_frame_get_filename(frame));
@@ -871,6 +899,11 @@ muse_processing_check_input(muse_processing *aProcessing, unsigned char aIFU)
   If one IFU of the same exposure is given multiple times in the input
   frameset, only the first one is used and a warning is printed for all others.
 
+  @note This function also modifies the usedframes component of the aProcessing
+        structure, and adds all newly found exposures to that frameset. However,
+        it is possible to run this function more than once, since no duplicate
+        frames are added to that frameset.
+
   @error{set CPL_ERROR_NULL_INPUT\, return NULL, invalid processing pointer}
   @error{set CPL_ERROR_DATA_NOT_FOUND\, return NULL, no exposures found}
  */
