@@ -134,6 +134,8 @@ muse_utils_replace_nan(muse_image *aImage, float aValue)
   return CPL_ERROR_NONE;
 }
 
+/* XXX: Moved to muse_utils, kept here disabled until final place is found */
+#if 0
 /*----------------------------------------------------------------------------*/
 /**
   @brief    Load a FOV image into a MUSE image.
@@ -142,7 +144,7 @@ muse_utils_replace_nan(muse_image *aImage, float aValue)
 
   Load the DATA extension containing the FOV image data from the input
   file. Bad pixels may be encoded as NaN values in the image data directly,
-  or, if a DQ extension is found in the inputfile, it is used to replace
+  or, if a DQ extension is found in the input file, it is used to replace
   the bad pixels in the FOV image data with NaN values.
 
   Alternatively, if the DATA extension is an extension with higher-dimensional
@@ -269,6 +271,7 @@ muse_fov_load(const char *aFilename)
   cpl_free(extname);
   return image;
 }
+#endif
 
 /* XXX: Maybe there is a better name for this function */
 /*----------------------------------------------------------------------------*/
@@ -634,16 +637,14 @@ muse_align_estimate_offsets(double *aOffsetRA, double *aOffsetDEC, double *aWeig
                             const cpl_matrix *aDeltaRA, const cpl_matrix *aDeltaDEC,
                             const cpl_matrix *aMeanDEC, double aRadius, int aNbins)
 {
-  cpl_ensure((aOffsetRA && aOffsetDEC) && aWeight, CPL_ERROR_NULL_INPUT,
-             CPL_FALSE);
-  cpl_ensure((aDeltaRA && aDeltaDEC) && aMeanDEC, CPL_ERROR_NULL_INPUT,
-             CPL_FALSE);
+  cpl_ensure((aOffsetRA && aOffsetDEC) && aWeight, CPL_ERROR_NULL_INPUT, 0);
+  cpl_ensure((aDeltaRA && aDeltaDEC) && aMeanDEC, CPL_ERROR_NULL_INPUT, 0);
   cpl_ensure(cpl_matrix_get_nrow(aDeltaRA) == cpl_matrix_get_nrow(aDeltaDEC),
-             CPL_ERROR_INCOMPATIBLE_INPUT, CPL_FALSE);
+             CPL_ERROR_INCOMPATIBLE_INPUT, 0);
   cpl_ensure(cpl_matrix_get_nrow(aDeltaRA) == cpl_matrix_get_nrow(aMeanDEC),
-             CPL_ERROR_INCOMPATIBLE_INPUT, CPL_FALSE);
-  cpl_ensure(aRadius > 0., CPL_ERROR_ILLEGAL_INPUT, CPL_FALSE);
-  cpl_ensure(aNbins > 0, CPL_ERROR_ILLEGAL_INPUT, CPL_FALSE);
+             CPL_ERROR_INCOMPATIBLE_INPUT, 0);
+  cpl_ensure(aRadius > 0., CPL_ERROR_ILLEGAL_INPUT, 0);
+  cpl_ensure(aNbins > 0, CPL_ERROR_ILLEGAL_INPUT, 0);
 
   cpl_matrix *distance = muse_align_compute_distances(aDeltaRA, aDeltaDEC,
                                                       aMeanDEC, NULL, NULL);
@@ -662,9 +663,20 @@ muse_align_estimate_offsets(double *aOffsetRA, double *aOffsetDEC, double *aWeig
   cpl_matrix *dDEC = muse_cplmatrix_extract_selected(aDeltaDEC, selection);
   cpl_array_delete(selection);
 
-  /* Search box from -aRadius to aRadius for both axes */
+  /* Search box from -aRadius to aRadius for both axes. */
   double bstep = 2. * aRadius / (double)aNbins;
 
+  /* Note: Despite of the previous selection of positions which have        *
+   * distances smaller than the search radius the bin index of the right    *
+   * ascension may become invalid and must be restricted to valid range.    *
+   * The reason is that in the following the difference in right ascension  *
+   * is directly compared to the search radius, while in the selection of   *
+   * positions this difference is corrected by the cosine of the mean       *
+   * declination (cf. muse_align_compute_distances()), and thus differences *
+   * in right ascension which are larger than the search radius may still   *
+   * be accepted by the distance criterion!                                 *
+   * To play it save also the index along the y-axis is validated.          */
+
   cpl_matrix *bins = cpl_matrix_new(1, aNbins);
   cpl_size ibin;
   for (ibin = 0; ibin < aNbins; ++ibin) {
@@ -673,15 +685,25 @@ muse_align_estimate_offsets(double *aOffsetRA, double *aOffsetDEC, double *aWeig
   cpl_matrix *histogram = cpl_matrix_new(aNbins, aNbins);
   double *hdata = cpl_matrix_get_data(histogram);
   cpl_size npos = cpl_matrix_get_ncol(dRA);
+  cpl_size nmatch = 0;
   cpl_size ipos;
   for (ipos = 0; ipos < npos; ++ipos) {
     cpl_size ix = (cpl_size)((cpl_matrix_get(dRA, 0, ipos) + aRadius) / bstep);
     cpl_size iy = (cpl_size)((cpl_matrix_get(dDEC, 0, ipos) + aRadius) / bstep);
-    hdata[iy * aNbins + ix] += 1.;
+    if (((ix >= 0) && (ix < aNbins)) && ((iy >= 0) && (iy < aNbins))) {
+      hdata[iy * aNbins + ix] += 1.;
+      ++nmatch;
+    }
   }
   cpl_matrix_delete(dRA);
   cpl_matrix_delete(dDEC);
 
+  if (nmatch == 0) {
+    cpl_matrix_delete(histogram);
+    cpl_matrix_delete(bins);
+    return 0;
+  }
+
   cpl_size row;
   cpl_size column;
   cpl_matrix_get_maxpos(histogram, &row, &column);
@@ -692,7 +714,7 @@ muse_align_estimate_offsets(double *aOffsetRA, double *aOffsetDEC, double *aWeig
   cpl_matrix_delete(histogram);
   cpl_matrix_delete(bins);
 
-  return npos;
+  return nmatch;
 }
 
 /*----------------------------------------------------------------------------*/
@@ -1032,6 +1054,28 @@ muse_align_compute_field_offsets(muse_imagelist *aImagelist,
   cpl_table_subtract_scalar(offsets, MUSE_OFFSETS_DRA, ra_shift);
   cpl_table_subtract_scalar(offsets, MUSE_OFFSETS_DDEC, dec_shift);
   cpl_table_set_column_invalid(offsets, MUSE_OFFSETS_FSCALE, 0, nfields);
+
+  /* Extra QC parameters: offset statistics */
+  double qcmin  = cpl_table_get_column_min(offsets, MUSE_OFFSETS_DRA);
+  double qcmax  = cpl_table_get_column_max(offsets, MUSE_OFFSETS_DRA);
+  double qcmean = cpl_table_get_column_mean(offsets, MUSE_OFFSETS_DRA);
+  double qcsdev = cpl_table_get_column_stdev(offsets, MUSE_OFFSETS_DRA);
+
+  cpl_propertylist_append_double(aHeader, QC_ALIGN_DRA_MIN, qcmin);
+  cpl_propertylist_append_double(aHeader, QC_ALIGN_DRA_MAX, qcmax);
+  cpl_propertylist_append_double(aHeader, QC_ALIGN_DRA_MEAN, qcmean);
+  cpl_propertylist_append_double(aHeader, QC_ALIGN_DRA_STDEV, qcsdev);
+
+  qcmin  = cpl_table_get_column_min(offsets, MUSE_OFFSETS_DDEC);
+  qcmax  = cpl_table_get_column_max(offsets, MUSE_OFFSETS_DDEC);
+  qcmean = cpl_table_get_column_mean(offsets, MUSE_OFFSETS_DDEC);
+  qcsdev = cpl_table_get_column_stdev(offsets, MUSE_OFFSETS_DDEC);
+
+  cpl_propertylist_append_double(aHeader, QC_ALIGN_DDEC_MIN, qcmin);
+  cpl_propertylist_append_double(aHeader, QC_ALIGN_DDEC_MAX, qcmax);
+  cpl_propertylist_append_double(aHeader, QC_ALIGN_DDEC_MEAN, qcmean);
+  cpl_propertylist_append_double(aHeader, QC_ALIGN_DDEC_STDEV, qcsdev);
+
   return offsets;
 }
 
